//--------------------------------------------------------------------
// FloppyTon II (A TON of floppies for Apple II)
//  Jesús Arias (2023)
//  GUS16-V5 CPU + external byte swapper
//--------------------------------------------------------------------

`include "cpuV5.v"

module floppygus (
	input clk,		// 25MHz
	input reset,
	
	// VIDEO
	input hsyn,		// VGA sync input
	input vsyn,		// VGA sync input
	output video,	// 1-bit video out
	output rem,		// video remark
	output viden,	// video enable (with blank)
	// SPI
	input	miso,
	output	sck,
	output	mosi,
	output	cs2b,	// SD card
	// PS2 KEYBOARD
	input	kclk,
	input	kdat,
	output 	grab,	// Take control of keyboard
	// disk
	input [3:0]phs,	// stepper motor phases
	input drvon,	// motor ON
	input drvsel,	// disk 1 / disk 2
	output wrprot,	// write protected
	input wrrq,		// write request
	input wrdata,	// Write data stream	
	output rddata,	// Read data stream
	// flow control
	output waitwr,	// not ready for writes
	// UART
	output txd,
	input rxd,
	
	output debug
);

/////////////////////////////// MEMORY ////////////////////////////////

wire [15:0]cdi;	// CPU Data bus, Input
wire [15:0]cdo;	// CPU Data bus, Output
wire [15:0]ca;	// CPU Address bus

wire [15:0]mdi;	// Memory data input
wire [15:0]mdo8;	// Memory data output
wire [15:0]mdo4;	// Memory data output
wire [12:0]ma;	// memory address
wire mwe;		// memory Write enable
assign mwe=((~vrd)&(~dmagr)& we) | dmawe;
assign ma=vrd ? va[13:1] : (dmagr ? dmaaddr : ca[12:0]);
assign mdi= dmagr ? dmawrdata : cdo; 

// writes allowed only to not-code areas: (0x040 - 0x07F), (0x900 - 0xFFF)
wire writable8 = (ma[11:6]==6'b000001) | (ma[11]&(ma[10]|ma[9]|ma[8]));
genram #(
`ifdef SIMULATION
	.INITFILE("romGUS.hex"),
`else
	.INITFILE("rand.hex"),
`endif
	.AW(12),	// 4Kword
	.DW(16)
	) ram8 (.clk(~clk), .we(mwe&(~ma[12])&writable8), .addr(ma[11:0]),.data_out(mdo8),.data_in(mdi));
// high RAM always writable	
genramx #(
	.AW(11),	// 2Kword
	.DW(16)
	) ram4 (.clk(~clk), .we(mwe&ma[12]&(~ma[11])), .addr(ma[10:0]),.data_out(mdo4),.data_in(mdi));
wire [15:0]mdo = ma[12]?mdo4:mdo8;

wire [7:0]vmdo = va[0]?mdo[7:0]:mdo[15:8]; // video big endian

//////////////////////////////////////////////////
//////					CPU					//////
//////////////////////////////////////////////////

reg [3:0]resc=4'b0001; //1111;
always @(posedge hsyn) if (resc!=0) resc<=resc-1;
//wire resetb=~((resc!=0) | reset);
wire resetx=((resc!=0) | reset);

wire cclk=clk|vrd|dmagr;

assign cdi = iocs ? iodo : (ca[15]? {mdo[7:0],mdo[15:8]}:mdo);

wire we;

//CPU_1CV2 #(.VECTORBASE(16'h2000)) GUScpu0( .clkin(cclk), .we(we), .a(ca), .Do(cdo), .di(cdi), 
//	.waitin(1'b0), .resetb(resetb), .irq(irq), .ivector(ivector) );

CPU_GUSV5 #(.VECTORBASE(16'h2000)) GUScpu0( .clk(cclk), .we(we), .ca(ca), .cdo(cdo), .cdi(cdi), 
	 .reset(resetx), .irq(irq), .ivector(ivector) );

// IRQS
wire [2:0]ivector=3'b001;
wire irq;

// IRQEN reg
reg irqen=0;	
always @(posedge clk or posedge reset ) 
	if (reset) irqen<=0;
	else irqen <= iowe0 ? cdo[0] : irqen;
	
// phase rising edge interrupts
reg [3:0]phsold=0;
always @(posedge clk) phsold<=phs;
wire [3:0]phsrising=phs&(~phsold);
reg [3:0] phrlatched=0;
always @(posedge clk)
	if(clredges) phrlatched<=0;
	else phrlatched<=phrlatched|phsrising;

assign irq=(|phrlatched)&irqen;

/////////////////////////////////////////////////////////
///////////////////// Peripherals ///////////////////////
/////////////////////////////////////////////////////////

wire iocs;		// IO select ($0020 to $003F)

assign iocs=(ca[15:5]==11'b00000000001)&(~dmagr)&(~vrd);
wire iowe0,iowe1,iowe2,iowe3,iowe4,iowe5,iowe6,iowe7;	// Write strobes
assign iowe0=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&(~ca[1])&(~ca[0]);
assign iowe1=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&(~ca[1])&( ca[0]);
assign iowe2=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&(~ca[0]);
assign iowe3=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&( ca[0]);
assign iowe4=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&(~ca[0]);
assign iowe5=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&( ca[0]);
assign iowe6=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&( ca[1])&(~ca[0]);
assign iowe7=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&( ca[1])&( ca[0]);

// write strobes (data written ignored)
wire clredges = iowe1;	// Clear phase rising edges register (and interrupt)
wire clrdirty = iowe6;	// Clear DIRTY flag
// Read strobes
wire iore2;		// Clear UART data valid flag
wire iore5;		// Clear Keyboard KVAL flag
assign iore2=(~we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&(~ca[0]);
assign iore5=(~we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&( ca[0]);

/////////////////////////////
// I/O data to CPU MUX
reg [15:0]iodo;
always@*
  case (ca[2:0])
     0 : iodo <=  {15'h0,irqen}; // IRQ Enable
     1 : iodo <=  {dirty,11'h0,kval,spibusy,utxbusy,urxval}; // PFLAGS
     2 : iodo <=  {8'h0,urxbuffer}; // UART data
     3 : iodo <=  {drvsel,9'h0,control[5:0]};		// control
     4 : iodo <=  spibuffer;
     5 : iodo <=  {8'h0,keysc};
     6 : iodo <=  {drvon,11'h0,phrlatched};
     7 : iodo <=  16'hxxxx;
     default : iodo <= 16'hxxxx;
  endcase

////// Control reg //////

reg [5:0]control=0;
always @(posedge clk) if (iowe3) control<=cdo[5:0];
assign cs2b=control[0];
assign en_dma=control[1];
assign spifast=control[2];
assign grab=control[3];

wire [1:0]writeok=control[5:4];
assign wrprot = ~(drvon ?(writeok[drvsel]):0);

///////////////////// UART TX ////////////////////
// micro-UART RX, only 8N1, 115200 bps, clk=25MHz
reg [7:0]txdiv=0;
wire txdivmax=(txdiv==216);
always @(posedge clk) if (iowe2 | txdivmax) txdiv<=0; else txdiv<=txdiv+1;

reg [8:0]txsh=10'h1FF;
always @(posedge clk)
	if (iowe2) begin 
		`ifdef SIMULATION
	        $write ("%c",cdo[7:0]); $fflush ( );
		`endif
		 txsh<={cdo[7:0],1'b0}; 
	end	else if (txdivmax) txsh<={1'b1,txsh[8:1]};
assign txd = txsh[0];

reg [3:0]txbitcnt=0;
wire utxbusy=|txbitcnt;
always @(posedge clk) 
	if (iowe2) txbitcnt<=10; 
	else if (utxbusy&txdivmax) txbitcnt<=txbitcnt-1;
	
///////////////////// UART RX ////////////////////
// micro-UART RX, only 8N1, 115200 bps, clk=25MHz

reg [1:0]rxreg=0; // two samples of RXD
always @(posedge clk) rxreg<={rxreg[0],rxd};
reg [7:0]rxdiv=0;
always @(posedge clk)
	if ((rxreg[1]^rxreg[0])|(rxdiv==216)) rxdiv<=0; // Reset if max or on any RXD edge
	else rxdiv<=rxdiv+1;
wire rxsample = (rxdiv==108); // sample at the middle of the bit

reg [9:0]urxsh=10'h3FF; // 10-bit shift register
reg [7:0]urxbuffer=0;
reg urxval=0;
always @(posedge clk) begin
	if (~urxsh[0]) urxsh<=10'h3FF;  // if START bit at LSB set to all ones and copy results
	else if (rxsample) urxsh<={rxreg[0],urxsh[9:1]};
	
	if (~urxsh[0]) urxbuffer<=urxsh[8:1]; // remove START and STOP from the received data
	if (~urxsh[0]) urxval<=1;  // set data valid
	else if (iore2) urxval<=0;  // clear data valid flag on reads

end

///////////////////// SPI ////////////////////

reg [7:0]sdiv=0;
always @(posedge clk) sdiv<=wspi ? 0 : sdiv+1;
reg [15:0]spibuffer;
reg [15:0]shbusy=0;
wire spibusy;
wire spicke=(sdiv==8'hff)|spifast;
wire wspi=iowe4|iowe5;
reg spidt16=0;
always @(posedge clk) if (wspi) spidt16<=iowe4; 
assign spibusy=shbusy[15];
always @(posedge clk or posedge reset) begin
	if (reset) spibuffer<=0;
	else begin
		if (wspi) begin spibuffer<=cdo; shbusy<=iowe5? 16'hff00: 16'hffff; end
		else if (spicke&spibusy) begin
			spibuffer<=spidt16 ? {spibuffer[14:0],smiso}: {8'h00,spibuffer[6:0],smiso};
			shbusy<={shbusy[14:0],1'b0};
			end
	end
end
assign sck=spifast? (spibusy&(~clk)) : (spibusy&(sdiv[7]));
assign mosi=spidt16 ? spibuffer[15] : spibuffer[7];

//reg smiso;
//always @(posedge sck) smiso<=miso;
wire smiso=miso;

//////////////////// KEYBOARD //////////////////////
// PS2 serial receiver
////////////////////////////////////////////////////

reg [10:0]kreg=11'h7FF;	// shift register, 11-bit
reg krck=1;				// sampled keyboard clock
always @(posedge clk) krck<=kclk;	

reg kock=1;			// old clock sample (for edge detection)
wire newsc;			// new scancode received
wire [7:0]scancode;	
reg [7:0] keysc;
reg kval=0;			// key valid

assign newsc=~kreg[0];		// zero start bit => scancode received
assign scancode=kreg[8:1];

// shit reg.gets filled with ones after a key is received
always @(posedge clk) begin
	kock<=krck;
	kreg<=newsc ?11'h7FF : ( ((~krck)&kock) ? {kdat,kreg[10:1]} : kreg);
	if (newsc) keysc<=scancode;
	kval<=newsc? 1 : (iore5 ? 0: kval);
end

/////////////////////////////////////////////////////
//                    DMA
/////////////////////////////////////////////////////
// dividers
reg [4:0]div1=0;			// 1/24 (Tcpu)
wire tc1 = (div1>=23);		// terminal count 1 (every CPU cycle ~1us)
always @(posedge clk) div1<=(tc1) ? 0 : div1+1; 
reg [1:0]div2=0;			// 1/4 (Tdisk)
always @(posedge clk) if (tc1) div2<={~div2[0],div2[1]};
wire ck250 = (div2==2'b01);	// 25% duty
wire tc2 = tc1&div2[0]&div2[1];	// terminal count 2 (every disk bit ~4us)
reg [3:0]div3=0;			// 1/16 (Tdma)
always @(posedge clk) if (tc2) div3<=div3+1;
wire tc3=tc2&(div3==4'b1111);	// terminal count 3 (every disk word ~64us)

// dma request
reg dmarq=0;
always @(posedge clk) 
	if(tc3) dmarq<=1;
	else if (~vrd) dmarq<=0;
wire dmagr = dmarq&(~vrd)&en_dma;	

// shifting
reg [15:0]shbuf=0;
reg [15:0]shout=0;

always @(posedge clk) begin
	if (tc3&(~xwrrq)) shout<=shbuf; else if (tc2) shout<={shout[14:0],wrdbit};
	if (dmagr) shbuf<=mdo;	
end
// read stream (1us pulses for ones)
assign rddata=shout[15] & ck250 & en_dma & drvon;


// dma address for read (byte buffer: $1200 to $2BFF, or 6.5 KBytes)
// last 264 bytes not included (fine rpm tunning)

reg [12:0]dma=13'h900;
always @(posedge clk) 
	if (tc3) begin
		if (dma>=13'h157b) dma<=13'h900; else dma<=dma+1; // 300 rpm
	end
///////// dma write /////////
reg [1:0]lwrrq=0;	// Latched write request 
always @(posedge clk) if (tc3) lwrrq<=wrrq ? 2'b10 : ((lwrrq[1]|lwrrq[0])?lwrrq-1 : 0);
wire xwrrq=wrrq|lwrrq[1]|lwrrq[0];	// extended write request

reg wrdata1;	// NRZI to NRZ
always @(posedge clk) if (tc2) wrdata1<=wrdata;
wire wrdbit=wrdata^wrdata1;

reg [15:0]dmawrdata;		// latched DMA data (write)
always @(posedge clk) if (tc3) dmawrdata<=shout;

reg dwrq=0;				// DMA write request
always @(posedge clk) 
	if (tc3) dwrq<=drvon & (~wrprot) & xwrrq & en_dma;
	else if (dmawe) dwrq<=0;

wire dmawe = dwrq & dmagr; // DMA write enabled
reg dirty=0;			// Dirty flag
always @(posedge clk) 
	if (dmawe) dirty<=1;
	else if (clrdirty) dirty<=0;

// address mux

wire [12:0]dmaaddr = dma; 


////// Stop Apple II if writing while not ready 
assign waitwr = (~en_dma) & wrrq;
assign debug=en_dma;



////////////////////////////////////////////////////////////////////////
////////////////////////////// Video ///////////////////////////////////
// External hsyn, vsyn
// Text mode: 32x24, 128 chars
// Bit MSB of char: Remark (change colors or something)
////////////////////////////////////////////////////////////////////////
reg [9:0]hc=0;  // Horizontal counter
reg hblk=0;
always @(posedge clk) begin
        if (~hsyn) hc<=10'd930; else hc<=hc+1;
        if (hc==10'd511) hblk<=1;
        else if (hc==10'd1023) hblk<=0;
end
reg [9:0]vc=990;  // Vertical Counter
reg vblk=1;
always @(negedge hblk) begin
//        if (~vsyn) vc<=10'd942; else vc<=vc+1;
        if (~vsyn) vc<=10'd990; else vc<=vc+1;
        if (vc==10'd383) vblk<=1;
        else if (vc==10'd1023) vblk<=0;
end
wire blk;	// Video blank signal
assign blk=hblk|vblk;

// Video address, shifting, & related
wire vrd;	// Read video memory
wire [13:0]va;	// video address
wire load;		// Load shift register
wire shift;		// Shift pixels
reg [7:0]char;	// ASCII value
reg [7:0]sh;	// shift register
reg remark;		// video remark
reg opaque;		// Opaque/transp
reg dblk;		// Delayed blank 

assign vrd=(hc[3:1]==3'b000)&(~blk);
assign load=vrd&hc[0]; //(hc[3:0]==4'b0001)&(~blk);
assign shift=hc[0]; //&(~blk);
always @(posedge clk) begin
	if (vrd&(~load)) char<=vmdo;
	if (load) sh<=vmdo; else if (shift) sh<={sh[6:0],1'bx};
	if (load) remark<=char[7];
	if (load) opaque<=char[6]|char[5]; // ASCII<32 => Transparency
	if (hc[0]) dblk<=blk;
end
assign va=load ? {4'b0000,char[6:0],vc[3:1]} 	// Character pixels
			   : {4'b1011,vc[8:4],hc[8:4]};		// Text in RAM (11*1024)


assign video=sh[7]&(~dblk);
assign rem=remark&(~dblk);
assign viden=(~dblk)&opaque;


endmodule



//----------------------------------------------------------------------------
//-- Generic synchronous RAM memory
//----------------------------------------------------------------------------

module genram (        
    input clk,	// data out on rising edges
    input we,	// Write enable                  
    input wire [AW-1: 0] addr,      
    input wire [DW-1: 0] data_in,
    output reg [DW-1: 0] data_out   
);

parameter INITFILE = "rand.hex";
parameter AW = 13;
parameter DW = 8;

reg [DW-1: 0] mem [0: (1<<AW)-1];

// synchronous read
always @(posedge clk) begin
	data_out <= mem[addr];
end
// write
always @(posedge clk) begin
	if (we) mem[addr]<=data_in;
end

//-- Load contents with hex file
initial begin
  $readmemh(INITFILE, mem);
end

endmodule

module genramx (        
    input clk,	// data out on rising edges
    input we,	// Write enable                  
    input wire [AW-1: 0] addr,      
    input wire [DW-1: 0] data_in,
    output reg [DW-1: 0] data_out   
);

parameter AW = 13;
parameter DW = 8;

reg [DW-1: 0] mem [0: (1<<AW)-1];

// synchronous read
always @(posedge clk) begin
	data_out <= mem[addr];
end
// write
always @(posedge clk) begin
	if (we) mem[addr]<=data_in;
end

integer i;

initial begin
	for (i=0;i<(1<<AW); i=i+1) mem[i]=0; 
end
endmodule

