//--------------------------------------------------------------------
// FloppyTon II (A TON of floppies for Apple II)
//  Jesús Arias (2023)
//  Z80 CPU
//--------------------------------------------------------------------

module floppyton (
	input clk,		// 25MHz
	input reset,
	
	// VIDEO
	input hsyn,		// VGA sync input
	input vsyn,		// VGA sync input
	output video,	// 1-bit video out
	output rem,		// video remark
	output viden,	// video enable (with blank)
	// SPI
	input	miso,
	output	sck,
	output	mosi,
	output	cs2b,	// SD card
	// PS2 KEYBOARD
	input	kclk,
	input	kdat,
	output 	grab,	// Take control of keyboard
	// disk
	input [3:0]phs,	// stepper motor phases
	input drvon,	// motor ON
	input drvsel,	// disk 1 / disk 2
	output wrprot,	// write protected
	input wrrq,		// write request
	input wrdata,	// Write data stream	
	output rddata,	// Read data stream
	// flow control
	output waitwr,	// not ready for writes
	// for scope probes
	output debug
);

assign debug=en_dma;

/////////////////////////////// CPU ////////////////////////////////

wire reset_n,wait_n,int_n,nmi_n,busrq_n,cen;
wire m1_n,mreq_n,iorq_n,rd_n,wr_n,rfsh_n,halt_n,busak_n;
wire [7:0]cdi;	// CPU Data bus, Input
wire [7:0]cdo;	// CPU Data bus, Output
wire [15:0]ca;	// CPU Address bus

// always inactive signals
assign wait_n=1'b1;
assign nmi_n=1'b1;
assign busrq_n=1'b1;
assign cen=1'b1;		// Clock enable
wire cclk;
assign cclk=clk|vrd;
reg [3:0]resc=4'b0010;
always @(posedge hsyn) if (resc!=0) resc<=resc-1;
assign reset_n=~((resc!=0) | reset);

tv80s #(.Mode(1)) Z80CPU(
  // Outputs
  m1_n, mreq_n, iorq_n, rd_n, wr_n, rfsh_n, halt_n, busak_n, ca, cdo,
  // Inputs
  reset_n, cclk, wait_n, int_n, nmi_n, busrq_n, cdi, cen
  );
  
assign cdi=(~mreq_n) ? mdo : iodo;

wire [7:0]mdi;	// Memory data input
wire [7:0]mdo8;	// Memory data output
wire [7:0]mdo4;	// Memory data output
wire [13:0]ma;	// memory address
wire mwe;		// memory Write enable
assign mwe=((~vrd)&(~mreq_n)&(~wr_n))|dmawe;
assign ma=vrd ? va : (mreq_n? dma: ca[13:0]);
assign mdi=(~mreq_n) ? cdo : dmawrdata; 

genram #(
`ifdef SIMULATION
	.INITFILE("romFLOPPY.hex"),
`else
	.INITFILE("rand.hex"),
`endif
	.AW(13),	// 8KB
	.DW(8)
	) ram8 (.clk(~clk), .we(mwe&(~ma[13])), .addr(ma[12:0]),.data_out(mdo8),.data_in(mdi));
	
genramx #(
	.AW(12),	// 4KB
	.DW(8)
	) ram4 (.clk(~clk), .we(mwe&ma[13]&(~ma[12])), .addr(ma[11:0]),.data_out(mdo4),.data_in(mdi));
wire [7:0]mdo = ma[13]?mdo4:mdo8;


////////////////////////////////////////////////////////////////////////
////////////////////////////// Video ///////////////////////////////////
// External hsyn, vsyn
// Text mode: 32x24, 128 chars
// Bit MSB of char: Remark (change colors or something)
////////////////////////////////////////////////////////////////////////
reg [9:0]hc=0;  // Horizontal counter
reg hblk=0;
always @(posedge clk) begin
        if (~hsyn) hc<=10'd924; else hc<=hc+1;
        if (hc==10'd511) hblk<=1;
        else if (hc==10'd1023) hblk<=0;
end
reg [9:0]vc=990;  // Vertical Counter
reg vblk=0;
always @(negedge hblk) begin
//        if (~vsyn) vc<=10'd942; else vc<=vc+1;
        if (~vsyn) vc<=10'd990; else vc<=vc+1;
        if (vc==10'd383) vblk<=1;
        else if (vc==10'd1023) vblk<=0;
end
wire blk;	// Video blank signal
assign blk=hblk|vblk;

// Video address, shifting, & related
wire vrd;	// Read video memory
wire [13:0]va;	// video address
wire load;		// Load shift register
wire shift;		// Shift pixels
reg [7:0]char;	// ASCII value
reg [7:0]sh;	// shift register
reg remark;		// video remark
reg opaque;		// Opaque/transp
reg dblk;		// Delayed blank 

assign vrd=(hc[3:1]==3'b000)&(~blk);
assign load=vrd&hc[0]; //(hc[3:0]==4'b0001)&(~blk);
assign shift=hc[0]; //&(~blk);
always @(posedge clk) begin
	if (vrd&(~load)) char<=mdo;
	if (load) sh<=mdo; else if (shift) sh<={sh[6:0],1'bx};
	if (load) remark<=char[7];
	if (load) opaque<=char[6]|char[5]; // ASCII<32 => Transparency
	if (hc[0]) dblk<=blk;
end
assign va=load ? {4'b0000,char[6:0],vc[3:1]} 	// Character pixels
			   : {4'b1011,vc[8:4],hc[8:4]};		// Text in RAM (11*1024)


assign video=sh[7]&(~dblk);
assign rem=remark&(~dblk);
assign viden=(~dblk)&opaque;

/////////////////////////////////////////////////////
//                    DMA
/////////////////////////////////////////////////////
// dividers
reg [4:0]div1=0;			// 1/24 (Tcpu)
wire tc1 = (div1>=23);		// terminal count 1 (every CPU cycle ~1us)
always @(posedge clk) div1<=(tc1) ? 0 : div1+1; 
reg [1:0]div2=0;			// 1/4 (Tdisk)
always @(posedge clk) if (tc1) div2<={~div2[0],div2[1]};
wire ck250 = (div2==2'b01);	// 25% duty
wire tc2 = tc1&div2[0]&div2[1];	// terminal count 2 (every disk bit ~4us)
reg [2:0]div3=0;			// 1/8 (Tdma)
always @(posedge clk) if (tc2) div3<=div3+1;
wire tc3=tc2&(div3==3'b111);	// terminal count 3 (every disk byte ~32us)

// shifting
reg [7:0]shbuf=0;
reg [7:0]shout=0;
always @(posedge clk) begin
	if (tc3&(~xwrrq)) shout<=shbuf; else if (tc2) shout<={shout[6:0],wrdbit};
	if (mreq_n&(~vrd)) shbuf<=mdo;
end
// read stream (1us pulses for ones)
wire rddata=shout[7] & ck250 & en_dma & drvon;

// dma address for read (buffer: $1200 to $2BFF, or 6.5 KBytes)
// last 264 bytes not included (fine rpm tunning)
reg [13:0]dma=14'h1200;
always @(posedge clk) 
//	if (tc3) if (dma>=14'h2ADF) dma<=14'h1200; else dma<=dma+1;	// 301.1 rpm
	if (tc3) if (dma>=14'h2AF7) dma<=14'h1200; else dma<=dma+1; // 300 rpm
//	if (tc3) if (dma>=14'h2BFF) dma<=14'h1200; else dma<=dma+1;	// 288 rpm (full buffer)

///////// dma write /////////
reg [1:0]lwrrq=0;	// Latched write request 
always @(posedge clk) if (tc3) lwrrq<=wrrq ? 2'b10 : ((lwrrq[1]|lwrrq[0])?lwrrq-1 : 0);
wire xwrrq=wrrq|lwrrq[1]|lwrrq[0];	// extended write request

reg wrdata1;	// NRZI to NRZ
always @(posedge clk) if (tc2) wrdata1<=wrdata;
wire wrdbit=wrdata^wrdata1;

reg [7:0]dmawrdata;		// latched DMA data (write)
always @(posedge clk) if (tc3) dmawrdata<=shout;

reg dwrq=0;				// DMA write request
always @(posedge clk) 
	if (tc3) dwrq<=drvon & (~wrprot) & xwrrq & en_dma;
	else if (dmawe) dwrq<=0;

wire dmawe = dwrq & (~vrd) & mreq_n;	// DMA write enabled
reg dirty=0;			// Dirty flag
always @(posedge clk) 
	if (dmawe) dirty<=1;
	else if (clrdirty) dirty<=0;

////// Stop Apple II if writing while not ready 
assign waitwr = (~en_dma) & wrrq;

/////////////////////////////////////////////////////
/////////////////// PERIPHERALS /////////////////////
/////////////////////////////////////////////////////
wire wcontrol=(~iorq_n)&(~wr_n)&(~vrd)&(~ca[2])&(~ca[1])&(~ca[0]);
wire wspi=	  (~iorq_n)&(~wr_n)&(~vrd)&(~ca[2])&(~ca[1])&( ca[0]);
wire wdsk=	  (~iorq_n)&(~wr_n)&(~vrd)&(~ca[2])&( ca[1])&( ca[0]);

// strobes for flag clearing
wire rdkey=   (~iorq_n)&(~rd_n)&(~vrd)&(~ca[2])&( ca[1])&(~ca[0]);
wire clrphs=  (~iorq_n)&(~wr_n)&(~vrd)&( ca[2])&(~ca[1])&(~ca[0]);
wire clrdirty=(~iorq_n)&(~wr_n)&(~vrd)&( ca[2])&(~ca[1])&( ca[0]);


reg [4:0]control=4'hf;	// control I/O reg
reg [1:0]writeok=0;		// write ok on selected disk

// Output ports
always @(posedge clk or posedge reset) begin
	if (reset) control<=4'h7;
	else if(wcontrol) control<=cdo[7:4];
end
wire spifast,en_dma;
assign en_dma=control[1];
assign cs2b=control[0];
assign spifast=control[2];
assign grab=control[3];

always @(posedge clk or posedge reset) begin
	if (reset) writeok<=0;
	else if(wdsk) writeok<=cdo[1:0];
end
assign wrprot = ~(drvon ?(writeok[drvsel]):0);

// Input ports
reg [7:0]iodo;	// Peripherals data output (not a reg)
always @*
case (ca[1:0])
	2'b00:	iodo<={control[3:0],drvsel,drvon,sbusy,kval};
	2'b01:  iodo<=ssh;
	2'b10:	iodo<=keysc;
	2'b11: 	iodo<={phrlatched,1'b0,dirty,writeok};
endcase

// phase rising edge interrupts
reg [3:0]phsold=0;
always @(posedge cclk) phsold<=phs;
wire [3:0]phsrising=phs&(~phsold);
reg [3:0] phrlatched=0;
always @(posedge cclk)
	if(clrphs) phrlatched<=0;
	else phrlatched<=phrlatched|phsrising;

assign int_n=~(phrlatched!=0);

////////////////////////////////////////////
//////////////////// SPI ///////////////////
////////////////////////////////////////////
reg [7:0]sdiv=0;
always @(negedge clk) sdiv<=wspi ? 0 : sdiv+1;
reg [7:0]ssh;
reg [7:0]shbusy=0;
reg sin;
wire spicke;
wire sbusy;
assign spicke=(sdiv==8'hff)|spifast;
assign sbusy=shbusy[7];
always @(negedge clk or posedge reset) begin
	if (reset) ssh<=0;
	else begin
		if (wspi) begin ssh<=cdo; shbusy<=8'hff; end
		else if (spicke&sbusy) begin
			ssh<={ssh[6:0],smiso};
			shbusy<={shbusy[6:0],1'b0};
			end
	end
end
assign sck=spifast? (sbusy&clk) : (sbusy&sdiv[7]);
assign mosi=ssh[7];

reg smiso;
always @(posedge sck) smiso<=miso;


//////////////////// KEYBOARD //////////////////////
// PS2 serial receiver
////////////////////////////////////////////////////

reg [10:0]kreg=11'h7FF;	// shift register, 11-bit
reg krck=1;				// sampled keyboard clock
always @(posedge clk) krck<=kclk;	

reg kock=1;			// old clock sample (for edge detection)
wire newsc;			// new scancode received
wire [7:0]scancode;	
reg [7:0] keysc;
reg kval=0;			// key valid

assign newsc=~kreg[0];		// zero start bit => scancode received
assign scancode=kreg[8:1];

// shit reg.gets filled with ones after a key is received
always @(posedge clk) begin
	kock<=krck;
	kreg<=newsc ?11'h7FF : ( ((~krck)&kock) ? {kdat,kreg[10:1]} : kreg);
	if (newsc) keysc<=scancode;
	kval<=newsc? 1 : (rdkey? 0: kval);
end

endmodule



//----------------------------------------------------------------------------
//-- Generic synchronous RAM memory
//----------------------------------------------------------------------------

module genram (        
    input clk,	// data out on rising edges
    input we,	// Write enable                  
    input wire [AW-1: 0] addr,      
    input wire [DW-1: 0] data_in,
    output reg [DW-1: 0] data_out   
);

parameter INITFILE = "rand.hex";
parameter AW = 13;
parameter DW = 8;

reg [DW-1: 0] mem [0: (1<<AW)-1];

// synchronous read
always @(posedge clk) begin
	data_out <= mem[addr];
end
// write
always @(posedge clk) begin
	if (we) mem[addr]<=data_in;
end

//-- Load contents with hex file
initial begin
  $readmemh(INITFILE, mem);
end

endmodule

module genramx (        
    input clk,	// data out on rising edges
    input we,	// Write enable                  
    input wire [AW-1: 0] addr,      
    input wire [DW-1: 0] data_in,
    output reg [DW-1: 0] data_out   
);

parameter AW = 13;
parameter DW = 8;

reg [DW-1: 0] mem [0: (1<<AW)-1];

// synchronous read
always @(posedge clk) begin
	data_out <= mem[addr];
end
// write
always @(posedge clk) begin
	if (we) mem[addr]<=data_in;
end

integer i;

initial begin
	for (i=0;i<(1<<AW); i=i+1) mem[i]=0; 
end
endmodule

