;---------------------------------------------------------------------
;                    FloppyGUS-II source code
;                       Jesús Arias (2023)
;                   GUS16-V6 + HP-150 style font 
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; Programming Conventions:
; Strict:
; - Core revision 6 required 
; - R7 is the Stack Pointer
; - R6 holds the return address for subroutines
; - Full-descending Stack
; Loose guidelines (see specific function comments):
; - R0, R1, R2 are used for arguments / results
;              Values may change on subroutine calls
; - R3, R4, R5 are used for local variables
;              Values are preserved on subroutine calls
;---------------------------------------------------------------------
;---------------------------------------------------------------------
;----------- I/O MAP  -----------

IOBASE=		0x20
IRQEN=		0x20
PFLAGS=		0x21
CLREDGE=	0x21
UARTDAT=	0x22
CONTROL=	0x23
SPIDAT16=	0x24
SPIDAT=		0x25
KEYB=		0x25
FLAGS2=		0x26
CLRDIRTY=	0x26

; Offsets relative to IOBASE
OIRQEN=		0
OPFLAGS=	1
OCLREDGE=	1
OUARTDAT=	2
OCONTROL=	3
OSPIDAT16=	4
OSPIDAT=	5
OKEYB=		5
OFLAGS2=	6
OCLRDIRTY=	6

;--------- Interrupts --------
IRQ0EN=		1
;--------- PFLAGS register bits -----------
UARTDV=		1		; Data valid (UART RX)
UARTBUSY=	2		; Transmitter busy (UART TX)
SPIBUSY=	4		; SPI busy
KVAL=		8		; Keyboard data valid
	;		bit 15 (sign) is DIRTY

;--------- CONTROL register bits -----------
CSB=		1		; SD card Chip Select line
EN_DMA=		2		; Enable DMA
SPIFAST=	4		; SPI clock (1 = clk, 0 = clk/256)
GRAB=		8		; Grab computer keyboard
WROK1=		16		; Write enable drive 1
WROK2=		32		; Write enable drive 2
	;		bit 15 (sign) is DRVSEL
;--------- FLAGS2 register bits -----------
PHSMSK=		0xF		; phase rising edge bits
	;		bit 15 (sign) is DRVON

;--------------------------------------
;----------- VECTORS -----------------
;-------------------------------------- 

		org 0			; RESET
inicio:	jr		start

;--------------------------------------
;--------- Interrupt Vectors ----------
;-------------------------------------- 
; Actual jump to address 0x2004 on interrupt (core parameter)
; no I/O registers at that address, position independent code

		org		0x4		; IRQ 0
;		jr		irq0
;--------------------------------------
;-----------   Interrupts -------------
;-------------------------------------- 

irq0:	subi	r7,3
		st		(r7+0),r0
		st		(r7+1),r1
		st		(r7+2),r2
		
		ldi		r2,trk1
		ldi		r1,IOBASE
		ld		r0,(R1+OCONTROL)	; drive 1 or 2?
		jpl		ir2
		ldi		r2,trk2
ir2:	ld		r0,(r1+OFLAGS2)		; rising edge of phases
		st		(r1+OCLREDGE),r0	; Clear edge detector reg
		andi	r0,0xf
		or		r1,r0,r0
		
		andi	r1,0xA				; phs1 or psh3 ?
		jz		ir3					; even phase: increment or decrement
		ldi		r1,oddphs
		st		(r1),r0				; odd phase: store rising phases
		jr		ir9
		
ir3:	ldi		r1,oddphs
		ld		r1,(r1)				; oddphase | current
		or		r0,r0,r1
		cmpi	r0,0x9				; phs3 + phs0 -> trk++
		jz		irinc
		cmpi	r0,0x6				; phs1 + phs2 -> trk++
		jz		irinc
		cmpi	r0,0x3				; phs1 + phs0 -> trk--
		jz		irdec
		cmpi	r0,0xc				; phs3 + phs2 -> trk--
		jnz		ir9
irdec:	ld		r0,(r2)				; if (trk) trk--
		jz		ir9
		subi	r0,1
		jr		ir8
irinc:	ld		r0,(r2)				; trk++
		addi	r0,1
ir8:	st		(r2),r0

ir9:	ld		r0,(r7+0)			; return from interrupt
		ld		r1,(r7+1)
		ld		r2,(r7+2)
		addi	r7,3
		reti

		
;--------------------------------------
;----------- VARIABLES (near) ---------
;--------------------------------------
; Located below 0x100 in order to load their addresses with LDI

		org	0x40		
varbas:			; variables
oddphs:	word 0	; old rising phases (for track change)
trk1:	word 34	; current track for disk 1
trk2:	word 34	; current track for disk 2
otrk:	word 0	; old track (for track change detection)
okey:	word 0	; old keycode (for released keys)

fatbas:		; FAT variables
storage_begin:	
		word 0
		word 0
fat1_begin:	
		word 0
		word 0
root_cluster:
		word 0
		word 0

last_cluster:
		word 0
		word 0
l2spc:	word 0		; log2(sectors_per_cluster)
spc:	word 0		; sectors_per_cluster

		; File variables
filelen:
		word 0
		word 0
filecluster:
		word 0
		word 0
filesector:
		word 0
;bufix:	word 0	; for file streaming
;bufmax:	word 0	

; File Selector variables

fselbas:	
fsel:	word 0	; index of selection (<32)
pos:	word 0	; file position after offset
ppos:	word 0	; file position before offset
offset:	word 0	; first file displayed
curdir:	word 0 	; current directory cluster
		word 0
scr:	word 0	; scratch data

; Disk image variables

rw1:	word	0x524f	; RO
		word	0x313a	; 1:
info1:	word 0	; filename 1
		word 0
		word 0
		word 0
secdsk1:			; sector 1
		word 0
		word 0

rw2:	word	0x524f	; RO
		word	0x323a	; 2:
info2:	word 0	; filename 2
		word 0
		word 0
		word 0
secdsk2:			; sector 2
		word 0
		word 0
osector:	; old track sector (for write)
		word 0
		word 0

;--------------------------------------
;----------- 64 words of code ---------
;-------------------------------------- 
		org 4*32
;-----------------------------------------------
;       Clear screen
;-----------------------------------------------
; Parameters:	R0: fill word (2 characters)
; Returns:		none
; Modiffies:	R1, R2, vpointer=video_base (top-left)

cls:	ldpc 	r1
		word	video
		ldi		r2,192
cls1:	st		(r1),r0
		st		(r1+1),r0
		addi	r1,2
		subi	r2,1
		jnz		cls1
		
		jind	r6
;-----------------------------------------------
;       Draw a window frame
;-----------------------------------------------
; Parameters:	none
; Returns:		none
; Modiffies:	none

frameit:
		subi	r7,5
		st		(r7+0),r0
		st		(r7+1),r1
		st		(r7+2),r2
		st		(r7+3),r3
		st		(r7+4),r6
		
		ldpc	r0
		word	0x2020
		jal		cls
		
		ldpc	r0
		word	0xa0a0
		ldpc	r1
		word	0xa020
		ldpc	r2
		word	0x20a0
		ldpc	r3
		word	video
		ldi		r6,16
frm1:	st		(r3),r0
		addi	r3,1
		subi	r6,1
		jnz		frm1
		ldi		r6,22
frm2:	st		(r3),r1
		addi	r3,15
		st		(r3),r2
		addi	r3,1
		subi	r6,1
		jnz		frm2
		ldi		r6,16
frm3:	st		(r3),r0
		addi	r3,1
		subi	r6,1
		jnz		frm3

		ldpc	r1		; first text string
		word	txto1*2	
		ldpc	r2		; at line 2, column 2
		word	video*2+64+2
		ldi		r0,255	; end at string terminator
		jal		strncpy
		ldpc	r1		; next text string
		word	txto2*2	
		addi	r2,7	; at line 3, col 7
		ldi		r0,255	; end at string terminator
		jal		strncpy
		
		ld		r0,(r7+0)
		ld		r1,(r7+1)
		ld		r2,(r7+2)
		ld		r3,(r7+3)
		ld		r6,(r7+4)
		addi	r7,5
		jind	r6
	
;--------------------------------------
;----------- Character table  ---------
;-------------------------------------- 
; 6 words per character, big endian, starts with SPACE (ASCII 32)
		org		6*32
		;include "font150.hex"
		include "../HP150/HP150ROM.hex"

;--------------------------------------
;-----------   MAIN   -----------------
;-------------------------------------- 

start:

		ldpc	r7			; stack at the end of memory
		word	stack	
		ldi		r0,CSB		; all control bits inactive
		ldi		r1,IOBASE
		st		(r1+OCONTROL),r0
		st		(r1+OCLREDGE),r0
		st		(r1+OCLRDIRTY),r0

		ldi		r0,0		; hide screen
		jal		cls
		
		; clear image selection
		ldi		r1,rw1
		st		(r1+6),r0	; sect1=0
		st		(r1+7),r0
		st		(r1+14),r0	; sect2=0
		st		(r1+15),r0
		ldpc	r0
		word	0x524f		; RO
		st		(r1+0),r0
		st		(r1+8),r0
		ldpc	r0
		word	0x2020		; spaces
		ldi		r2,4
clrimg:	st		(r1+2),r0
		st		(r1+10),r0
		addi	r1,1
		subi	r2,1
		jnz		clrimg
		; init track counters
		ldi		r0,34
		ldi		r1,trk1
		st		(r1),r0
		st		(r1+1),r0
		
		ldi		r1,IOBASE	; enable interrupts
		ldi		r0,1
		st		(r1+OIRQEN),r0
		
		jal		waitaction

;--------------------------------------------------
;--------------------------------------
;  		File selection loop
;--------------------------------------
;--------------------------------------------------

;	ldi		r1,trk1
;	ldi		r5,10
;	ldi		r2,255
;tst1:	ld		r0,(r1)
;	sub		r6,r0,r2
;	jz		tst1
;	or		r2,r0,r0
;	jal		prthex
;	or		r2,r0,r0
;	jr		tst1


filesel:
		; card init
		jal		sd_init
		or		r0,r0,r0
		jnz		inie1
		jal		FAT_init
		or		r0,r0,r0
		jnz		inie2
		jr		fs000
inie1:	ldpc	r1
		word	2*txterrsd
		jr		inie3
inie2:	ldpc	r1
		word	2*txterrfat
inie3:	ldpc	r2
		word	2*video+8
		ldi		r0,16
		jal		bytecpy
		ldi		r6,16
		add		r2,r2,r6
		ldpc	r1
		word	2*txterrlin	
		jal		bytecpy
		jr		filesel

fs000:	; display init
		jal		frameit

		; start at root directory
		ldi		r2,fselbas
		ldi		r1,fatbas
		ld		r0,(r1+root_cluster-fatbas)
		st		(r2+curdir-fselbas),r0
		ld		r0,(r1+root_cluster-fatbas+1)
		st		(r2+curdir-fselbas+1),r0

fs00:	ldi		r2,fselbas
		ldi		r0,0	; fsel=pos=ppos=offset=0
		st		(r2+fsel-fselbas),r0
		st		(r2+pos-fselbas),r0
		st		(r2+ppos-fselbas),r0
		st		(r2+offset-fselbas),r0

fs0:	; Show current DISK images
		ldpc	r2			; line 5, col 2
		word	2*video+5*32+2
		ldi		r1,rw1*2
		ldi		r0,12
		jal		bytecpy
		addi	r2,3
		ldi		r1,rw2*2
		jal		bytecpy
		
		; read current directory
		
		ldi		r2,fselbas
		ldi		r1,fatbas
		ld		r0,(r2+curdir-fselbas)
		st		(r1+filecluster-fatbas),r0
		ld		r0,(r2+curdir-fselbas+1)
		st		(r1+filecluster-fatbas+1),r0
		ldi		r0,0
		st		(r1+filesector-fatbas),r0
		subi	r0,1
		st		(r1+filelen-fatbas),r0
		st		(r1+filelen-fatbas+1),r0

fs1:	; scan directory entries
		ldpc	r3
		word	fatbuf
		jal		rdfile
		ldpc	r1
		word	2*fatbuf

		ldi		r4,16
fs2:	jal		rdbyte
		or		r0,r0,r0
		jz		fs99
		cmpi	r0,0xE5		; deleted?
		jz		fsskip
		addi	r1,10
		jal		rdbyte
		subi	r1,12
		ldi		r6,0x30		; file or subdir
		and		r6,r0,r6
		jz		fsskip
		; File OK: count it
		ldi		r6,fselbas	; subdir attribute in scr
		andi	r0,0x10
		st		(r6+scr-fselbas),r0
		ld		r0,(r6+ppos-fselbas)
		ld		r2,(r6+offset-fselbas)
		sub		r2,r0,r2
		jz		fss1
		addi	r0,1
		st		(r6+ppos-fselbas),r0
		jr		fsskip
fss1:	; copy file metadata
		ld		r0,(r6+pos-fselbas)
		rori	r0,r0,12	; pos*16
		ldpc	r2
		word	filedata
		add		r6,r0,r2
		shr		r2,r1
		ldi		r5,16
fss11:	ld		r0,(r2)
		st		(r6),r0
		addi	r6,1
		addi	r2,1
		subi	r5,1
		jnz		fss11
		; print name, register file, register file
		ldi		r6,fselbas
		ld		r0,(r6+pos-fselbas)
		ld		r5,(r6+fsel-fselbas)
		ldi		r3,0
		sub		r5,r0,r5
		jnz		fss12
		ldi		r3,0x80
fss12:	jal		pos2video
		ldi		r0,8
		jal		bytecpyrem
		ldi		r0,' '
		jal		wrbyterem
		ldi		r0,3
		jal		bytecpyrem

		ldi		r0,' '
		jal		wrbyterem
		ldi		r6,fselbas
		ld		r5,(r6+scr-fselbas)
		or		r5,r5,r5
		jz		fss13
		ldi		r0,'D'
fss13:	jal		wrbyterem	
		subi	r1,11
		
fss2:
		ldi		r6,fselbas
		ld		r0,(r6+pos-fselbas)
		addi	r0,1
		st		(r6+pos-fselbas),r0
		cmpi	r0,32
		jz		fs99

fsskip:		; Next file entry
		addi	r1,32
		subi	r4,1
		jnz		fs2
		jr		fs1	; Read more sectors of directory

fs99:	; clear until the end of screen
		ldi		r5,fselbas
		ld		r0,(r5+pos-fselbas)
		or		r4,r0,r0
fs9900:	cmpi	r0,32
		jz		fs9903
		jal		pos2video
		ldi		r3,14
		ldi		r0,' '
fs99002:
		jal		wrbyte		
		subi	r3,1
		jnz		fs99002

		ld		r0,(r5+pos-fselbas)
		addi	r0,1
		st		(r5+pos-fselbas),r0
		jr		fs9900

fs9903:	st		(r5+pos-fselbas),r4
		;------------------------------------------------
		;-------------- Wait for KEY action -------------
		;------------------------------------------------
		jal		getkey
		cmpi	r0,0x72
		jz		kdown
		cmpi	r0,0x75
		jz		kup
		cmpi	r0,0x74
		jz		kright
		cmpi	r0,0x6B
		jz		kleft
		
		cmpi	r0,0x5A		;enter
		jz		kenterspc
		cmpi	r0,0x29		;space
		jz		kenterspc

		cmpi	r0,0x7c		;sys req
		jz		ksysrq
		cmpi	r0,0x16		;1
		jz		knum1
		cmpi	r0,0x1e		;2
		jz		knum2
	
		jr		fs999

		;------------------------------------------------
		;-------------- KEY actions -------------
		;------------------------------------------------

kup:	ld		r0,(r5+fsel-fselbas)
		jz		fs999
		subi	r0,1
		st		(r5+fsel-fselbas),r0
		jr		fs999

kdown:	ld		r0,(r5+fsel-fselbas)
		addi	r0,1
		st		(r5+fsel-fselbas),r0
		ld		r6,(r5+pos-fselbas)
		sub		r0,r0,r6
		jnc		fs999
		subi	r6,1
		st		(r5+fsel-fselbas),r6
		jr		fs999

kright:	ld		r0,(r5+fsel-fselbas)
		addi	r0,16
		ld		r1,(r5+pos-fselbas)
		sub		r1,r0,r1
		jc		fs9902
		st		(r5+fsel-fselbas),r0
		jr		fs999
fs9902:	cmpi	r0,32
		jc		fs999
		ld		r1,(r5+offset-fselbas)
		addi	r1,16
		st		(r5+offset-fselbas),r1
		jr		fs999

kleft:	ld		r0,(r5+fsel-fselbas)
		cmpi	r0,16
		jnc		fs9912
		subi	r0,16
		st		(r5+fsel-fselbas),r0
		jr		fs999

fs9912:	ld		r0,(r5+offset-fselbas)
		jz		fs999
		subi	r0,16
		st		(r5+offset-fselbas),r0
		jr		fs999


kenterspc:	; File or subdir selected
		or		r2,r0,r0
		ld		r0,(r5+fsel-fselbas)
		rori	r0,r0,12	; fsel*16
		ldpc	r1
		word	filedata
		add		r1,r0,r1	; r1: file info
		or		r4,r1,r1
		ld		r0,(r1+5)
		ldi		r6,0x10		; Subdirectory attributte
		and		r6,r0,r6
		jz		fs995
		; Subdirectory
		ld		r0,(r1+13)	; cluster LSB
		ld		r1,(r1+10)	; cluster MSB
		rori	r0,r0,8
		rori	r1,r1,8
		or		r6,r0,r1
		jz		filesel		; root cluster
		st		(r5+curdir-fselbas),r0
		st		(r5+curdir-fselbas+1),r1
		ldi		r6,filelen	; filelen = max.
		ldpc	r0
		word	0xffff
		st		(r6),r0
		st		(r6+1),r0
		jr		fs00

fs995:	ld		r0,(r1+4)	; file extension
		ldpc	r6
		word	0x5241		; "RA"
		sub		r6,r0,r6
		jnz		fs999
		ldi		r3,info1
		cmpi	r2,0x5A
		jz		f9951
		ldi		r3,info2	
f9951:	ldi		r6,4; copy file name
		ldpc	r2
		word	0x8080
f99510:	ld		r0,(r1)
		or		r0,r0,r2
		st		(r3),r0
		addi	r1,1
		addi	r3,1
		subi	r6,1
		jnz		f99510
		subi	r1,4

	;sector=storage_begin+((filecluster-2)<<l2spc)
		ld		r0,(r4+13)	; filecluster
		ld		r1,(r4+10)
		rori	r0,r0,8
		rori	r1,r1,8
		subi	r0,2
		sbci	r1,0
		
		ldi		r4,fatbas
		ld		r6,(r4+l2spc-fatbas)
f9952:	add		r0,r0,r0
		adc		r1,r1,r1
		subi	r6,1
		jnz		f9952
		ld		r2,	(r4+storage_begin-fatbas)
		ld		r6,	(r4+storage_begin-fatbas+1)
		add		r0,r0,r2
		adc		r1,r1,r6			
		st		(r3),r0
		st		(r3+1),r1

		ldi		r0,0xff
		ldi		r2,otrk
		st		(r2),r0
		jr		fs999
		
ksysrq:	jal		waitaction
		jal		frameit
		jr		fs999

knum1:	ldi		r4,WROK1
		ldi		r2,RW1
		jr		knum12
knum2:	ldi		r4,WROK2
		ldi		r2,RW2
knum12:	ldi		r1,IOBASE
		ld		r0,(r1+OCONTROL)
		xor		r0,r0,r4
		st		(r1+OCONTROL),r0
		ldpc	r6
		word	0x5257	; RW
		and		r4,r0,r4
		jnz		knum13
		ldpc	r6
		word	0x524F	; RO
knum13:	st		(r2),r6
		;jr		fs999:


fs999:	ldi		r0,0
		ldi		r5,fselbas
		st		(r5+pos-fselbas),r0
		st		(r5+ppos-fselbas),r0
		jr		fs0
		
;--------------------------------------------------
;--------------------------------------
;  		Floppy emulation loop
;--------------------------------------
;--------------------------------------------------
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;	
; Pops to foreground if 
;  <sysreq> is pressed
;  does track reading meanwhile
waitaction:
		subi	r7,4
		st		(r7+0),r3
		st		(r7+1),r4
		st		(r7+2),r5
		st		(r7+3),r6
		ldi		r1,IOBASE
		ld		r0,(r1+OCONTROL)	; ungrab keyboard
		andi	r0,0x3F^GRAB
		st		(r1+OCONTROL),r0	; enable interrupt
		ldi		r0,1
		st		(r1+OIRQEN),r0
		ldi		r0,0
		jal		cls

wa1:	ldi		r1,IOBASE
		ld		r0,(r1+OPFLAGS)
		andi	r0,KVAL
		jz		wa2
		ldi		r2,varbas
		ld		r3,(r2+okey-varbas)
		ld		r0,(r1+OKEYB)
		st		(r2+okey-varbas),r0
		cmpi	r0,0x7c	; sys req key
		jnz		wa2
		cmpi	r3,0xF0	; released key
		jz		wa2
		ld		r0,(r1+OCONTROL)
		ori		r0,GRAB
		st		(r1+OCONTROL),r0
		ld		r3,(r7+0)	
		ld		r4,(r7+1)	
		ld		r5,(r7+2)	
		ld		r6,(r7+3)	
		addi	r7,4
		jind	r6
				
wa2:	ld		r0,(r1+OFLAGS2)	; DRVON?
		jmi		wa20
		ld		r0,(r1+OPFLAGS)	; DIRTY?
		jpl		wa1
		; DIRTY was set
		ld		r0,(r1+OCONTROL)	
		andi	r0,0x3F^EN_DMA
		st		(r1+OCONTROL),r0
		jal		wrsectors
		jr		wa1
wa20:
		ldi		r2,varbas
		ld		r3,(r2+otrk-varbas)
		jal		gettrack
		sub		r6,r0,r3
		jz		wa1
		; track changed
wa21:
		ldi		r2,varbas
		st		(r2+otrk-varbas),r0
		ldi		r1,IOBASE
		ld		r4,(r1+OCONTROL)
		andi	r4,0x3F^EN_DMA
		st		(r1+OCONTROL),r4
		ld		r4,(r1+OPFLAGS)	; DIRTY?
		jpl		wa210
		jal		wrsectors
		
		; off = trk*13
wa210:	ldi		r2,varbas
		ld		r0,(r2+otrk-varbas)	
		andi	r0,0x7F
		rori	r3,r0,14	; r3=r0<<2
		rori	r4,r0,13	; r4=r0<<3
		add		r0,r0,r3
		add		r0,r0,r4	; r0=trk+trk*4+trk*8
		ldi		r2,secdsk1
		ldi		r1,IOBASE
		ld		r6,(r1+OCONTROL)	; drvsel?
		jpl		wa22
		ldi		r2,secdsk2
wa22:	ld		r6,(r2)		; R1:R2 sector = secdsk + off
		add		r1,r0,r6
		ld		r2,(r2+1)
		adci	r2,0		
		ldi		r4,osector
		st		(r4),r1
		st		(r4+1),r2
		ldpc	r3
		word	dmabuf
		ldi		r5,13
wa3:	jal		sd_rd_block
		addi	r1,1
		adci	r2,0
		jal		gettrack
		ldi		r6,otrk
		ld		r6,(r6)
		sub		r6,r0,r6
		jnz		wa21
		subi	r5,1
		jnz		wa3	

		ldi		r1,IOBASE
		ld		r0,(r1+OCONTROL)
		ori		r0,EN_DMA
		st		(r1+OCONTROL),r0
		
		jr		wa1

; Writting back the track
wrsectors:
		subi	r7,3
		st		(r7+0),r3
		st		(r7+1),r5
		st		(r7+2),r6

		ldpc	r3			; write mark
		word	dmabuf+3328-8-(endtxt-txttrk)
		ldpc	r5			
		word	txttrk
		ldi		r6,endtxt-txttrk
wrsct0:	ld		r1,(r5)
		st		(r3),r1
		addi	r5,1
		addi	r3,1
		subi	r6,1
		jnz		wrsct0
		
		ldi		r6,osector
		ld		r1,(r6)
		ld		r2,(r6+1)
		ldpc	r3
		word	dmabuf
		ldi		r5,13
wrsct1:	jal		sd_wr_block
		addi	r1,1
		adci	r2,0
		subi	r5,1
		jnz		wrsct1

		ldi		r6,IOBASE			; clear dirty flag
		st		(r6+OCLRDIRTY),r6

		ld		r3,(r7+0)
		ld		r5,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6

; returns current track
gettrack:
		subi	r7,1
		st		(r7),r6
		ldi		r6,IOBASE
		ld		r6,(r6+OCONTROL)	; DRVSEL is bit #15
		jpl		gtrk1
		ldi		r0,trk2
		ld		r0,(r0)
		ldi		r6,0x80		; disk 2 tracks marked
		or		r0,r0,r6
		jr		gtrk2
gtrk1:	ldi		r0,trk1
		ld		r0,(r0)	
gtrk2:	ld		r6,(r7)
		addi	r7,1
		jind	r6

;--------------------------------------
;  		SD card functions
;--------------------------------------
; CS high/Low
; Parameters:	none
; Returns:		none
; Modiffies:	no register
sd_cs_h:
		subi	r7,2
		st		(r7+0),r0
		st		(r7+1),r6
		ldi		r6,IOBASE
		ld		r0,(r6+OCONTROL)
		ori		r0,CSB
		jr		sd_cs1
		
sd_cs_l:
		subi	r7,2
		st		(r7+0),r0
		st		(r7+1),r6
		ldi		r6,IOBASE
		ld		r0,(r6+OCONTROL)
		andi	r0,0x3F^CSB
sd_cs1:	st		(r6+OCONTROL),r0
		ld		r0,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;--------------------------------------
; SD command
;
; Parameters:	R0: command
;				R2:R1 argument (32 bits)
; Returns:		R0: r1 state
; Modiffies:	nothing else

sd_cmd:
		subi	r7,3
		st		(r7+0),r1
		st		(r7+1),r4
		st		(r7+2),r6
	
		jal		sd_cs_l
		ldi		r4,0x87		; CRC
		or		r0,r0,r0
		jnz		sdcm1
		ldi		r4,0x95		; CRC
sdcm1:	ldi		r6,0x40
		or		r0,r0,r6
		jal		SPIxfer
		or		r0,r2,r2	; arg MSBs
		jal		SPIxfer16
		or		r0,r1,r1	; arg LSBs
		jal		SPIxfer16	
		or		r0,r4,r4	; CRC
		jal		SPIxfer

		ldi		r1,255
		;ldpc	r1
		;word	0x7000
sdcm2:	ldi		r0,0xff
		jal	SPIxfer
		cmpi	r0,0xff
		jnz		sdcm3
		subi	r1,1
		jpl		sdcm2

sdcm3:	or		r4,r0,r0
retcmd:	ldi		r0,0xff
		jal		SPIxfer
		jal		sd_cs_h
		ldi		r0,0xff
		jal		SPIxfer
		or		r0,r4,r4
		
		ld		r1,(r7+0)
		ld		r4,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6

;--------------------------------------
; SD command, 32 bit result
;
; Parameters:	R0: command
;				R2:R1 argument (32 bits)
; Returns:		R0: r1-state
;				R2:R1 32 bit result (sometimes)
; Modiffies:	nothing else
sd_cmd32:
		subi	r7,3
		st		(r7+0),r1
		st		(r7+1),r4
		st		(r7+2),r6
	
		jal		sd_cs_l
		ldi		r4,0x87		; CRC
		or		r0,r0,r0
		jnz		sdcm31
		ldi		r4,0x95		; CRC
sdcm31:	ori		r0,0x40
		jal		SPIxfer
		or		r0,r2,r2	; arg MSBs
		jal		SPIxfer16
		or		r0,r1,r1	; arg LSBs
		jal		SPIxfer16	
		or		r0,r4,r4	; CRC
		jal		SPIxfer

		ldi		r1,255
sdcm32:	ldi		r0,0xff
		jal		SPIxfer
		cmpi	r0,0xff
		jnz		sdcm33
		subi	r1,1
		jpl		sdcm32

sdcm33:
		or		r4,r0,r0
		ldpc	r0
		word	0xffff
		jal		SPIxfer16
		or		r2,r0,r0
		ldpc	r0
		word	0xffff
		jal		SPIxfer16
		or		r1,r0,r0
		jr		retcmd

;--------------------------------------
; SD init
;
; Parameters:	none
; Returns:		R0: 0 OK
; Modiffies:	R1,R2

sd_init:
		subi	r7,2
		st		(r7+0),r3
		st		(r7+1),r6
		ldi		r1,IOBASE			; slow clock and CS high
		ld		r0,(r1+OCONTROL)
		andi	r0,0x3F^SPIFAST
		ori		r0,CSB
		st		(r1+OCONTROL),r0

		ldi		r3,10				; 80 clock pulses
sdini1:	ldi		r0,0xff
		jal		SPIxfer
		subi	r3,1
		jnz		sdini1

		jal		sd_cs_l

		ldi		r3,128
sdini2:	ldi		r0,0	; CMD go idle state
		ldi		r2,0	; arg MSB
		ldi		r1,0	; arg LSB
		jal		sd_cmd
		cmpi	r0,1
		jz		sdini3
		subi	r3,1
		jnz		sdini2
		ldi		r0,1	; Error:timeout
sdini99:
		ld		r3,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
		
sdini3:	ldi		r3,255
sdini35:
		ldi		r0,8	; SEND_IF_COND
		ldi		r2,0
		ldpc	r1
		word	0x01aa
		jal		sd_cmd32
		cmpi	r0,1
		jz		sdini40
		subi	r3,1
		jpl		sdini35

sdini40:
		ldi		r3,255
sdini4:	ldi		r0,55	; APP_CMD
		ldi		r2,0
		ldi		r1,0
		jal		sd_cmd
		ldi		r0,41	; SD_SEND_OP_COND,0x40000000
		ldpc	r2
		word	0x4000
		ldi		r1,0
		jal		sd_cmd
		cmpi	r0,0
		jz		sdini5
		subi	r3,1
		jpl		sdini4
		ldi		r0,2	; Error card initialization failed
		jr		sdini99

sdini5:	ldi		r3,255
sdini6:	ldi		r0,58	; Read OCR
		ldi		r2,0
		ldi		r1,0
		jal		sd_cmd32
		cmpi	r0,0
		jz		sdini9
		subi	r3,1
		jpl		sdini6
		ldi		r0,3	; Error read OCR
		jr		sdini99

sdini9:	ldi		r1,IOBASE
		ld		r0,(r1+OCONTROL)
		ori		r0,SPIFAST
		st		(r1+OCONTROL),r0
		ldi		r0,0
		jr		sdini99

;--------------------------------------
; SD read block
;
; Parameters:	R2:R1 :	nº sector
;				R3: 	destination address
; Returns:		R0: 0 OK, !=0 Error
; Modiffies:	R3 += 256

sd_rd_block:
		subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6
		ldi		r0,17	; READ SINGLE BLOCK
		jal		sd_cmd
		or		r0,r0,r0
		jnz		sdrd9

		jal		sd_cs_l
		ldpc	r1		; timeout
		word	0x300
sdrd2:	ldi		r0,0xff
		jal		SPIxfer
		cmpi	r0,0xFE
		jz		sdrd3
		subi	r1,1
		jnz		sdrd2
		ldi		r0,4
		jr		sdrd9

sdrd3:	ldi		r1,255	; 256 words
sdrd4:	ldpc	r0
		word	0xffff
		jal		SPIxfer16
		st		(r3),r0
		addi	r3,1
		subi	r1,1
		jpl		sdrd4

		ldi		r0,0xff	; discard results and CRC
		jal		SPIxfer
		ldi		r0,0xff
		jal		SPIxfer
		ldi		r0,0xff
		jal		SPIxfer
		ldi		r0,0	; OK

sdrd9:	jal		sd_cs_h
		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6


;--------------------------------------
; SD write block
;
; Parameters:	R2:R1 :	nº sector
;				R3: 	source address
; Returns:		R0: 0 OK, !=0 Error
; Modiffies:	R3 += 256	
sd_wr_block:
		subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6
		ldi		r0,24	; WRITE  SINGLE BLOCK
		jal		sd_cmd
		or		r0,r0,r0
		jnz		sdrd9

		jal		sd_cs_l		
		ldi		r0,0xFE	;Send start block token 0xfe
		jal		SPIxfer

		ldi		r1,255	; Send data
sdwr4:	ld		r0,(r3)
		addi	r3,1
		jal		SPIxfer16
		subi	r1,1
		jpl		sdwr4

		or		r0,r1,r1
		jal		SPIxfer16
	
		ldi		r1,0	; wait for data response
sdwr5:	ldi		r0,0xff
		jal		SPIxfer
		cmpi	r0,0xff
		jnz		sdwr6
		subi	r1,1
		jnz		sdwr5
		jr		sdwr91		; timeout
sdwr6:	andi	r0,0x1F		; AAA='010' - data accepted ;
		cmpi	r0,5	
		jnz		sdwr91		; error	

		ldi		r1,0	;Wait while card busy
sdwr7:	ldi		r0,0xff
		jal		SPIxfer
		or		r0,r0,r0
		jnz		sdwr92
		subi	r1,1
		jnz		sdwr7

sdwr91:	ldi		r1,1	; return error
sdwr93:	ldi		r0,0xff
		jal		SPIxfer
		jal		sd_cs_h
		jal		SPIxfer
		or		r0,r1,r1
		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
sdwr92:	ldi		r1,0	; return OK
		jr		sdwr93

;--------------------------------------
;--------------------------------------
; FAT init
; Parameters:	none
; Returns:	A: 0 OK, !=0 Error
; Modiffies:	lots

FAT_init:
		subi	r7,3
		st		(r7+0),r4
		st		(r7+1),r5
		st		(r7+2),r6
		ldi		r5,fatbas
		ldi		r2,0
		ldi		r1,0
		ldpc	r3
		word	fatbuf
		jal		sd_rd_block	; read partition table
		or		r0,r0,r0
		jnz		fini99
		subi	r3,1
		ld		r0,(r3)
		rori	r0,r0,8
		ldpc	r6
		word	0xAA55
		sub		r6,r0,r6
		jnz		fini99
		ldi		r6,28
		sub		r3,r3,r6
		ld		r1,(r3)		; starting sector of partition #1
		ld		r2,(r3+1)
		rori	r1,r1,8
		rori	r2,r2,8
		subi	r3,227
		jal		sd_rd_block	; superblock
		or		r0,r0,r0
		jnz		fini99

		ldpc	r6			; Sectors/Cluster
		word	0x100
		sub		r6,r3,r6
		ld		r0,(r6+6)
		andi	r0,0xff
		st		(r5+spc-fatbas),r0
		ldi		r6,0
fini1:	addi	r6,1
		shr		r0,r0
		jnz		fini1
		subi	r6,1
		st		(r5+l2spc-fatbas),r6
	
		ldpc	r6			
		word	0x100
		sub		r3,r3,r6
		ld		r0,(r3+7)	; reserved sectors
		rori	r0,r0,8
		add		r1,r1,r0
		adci	r2,0
		st		(r5+fat1_begin-fatbas),r1
		st		(r5+fat1_begin-fatbas+1),r2
	
		ld		r0,(r3+8)	; number of FATs
		rori	r0,r0,8
		ldi		r6,0x12
		add		r3,r3,r6
		ld		r4,(r3)		;sectors/FAT
		ld		r6,(r3+1)
		rori	r4,r4,8
		rori	r6,r6,8
fini3:	add		r1,r1,r4
		adc		r2,r2,r6
		subi	r0,1
		jnz		fini3
		st		(r5+storage_begin-fatbas),r1
		st		(r5+storage_begin-fatbas+1),r2

		ld		r4,(r3+4)
		ld		r6,(r3+5)
		rori	r4,r4,8
		rori	r6,r6,8
		st		(r5+root_cluster-fatbas),r4
		st		(r5+root_cluster-fatbas+1),r6

		ldpc	r6
		word	0xffff
		st		(r5+last_cluster-fatbas),r6
		st		(r5+last_cluster-fatbas+1),r6
		
fini99:	ld		r4,(r7+0)
		ld		r5,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6
		
;--------------------------------------
;--------------------------------------
; next_cluster
; Parameters:	R2:R1 : current cluster
; Returns:		R0 : 0 if no error
;				R2:R1 : next cluster
; Modiffies:	nothing else

next_cluster:
		subi	r7,4
		st		(r7+0),r3
		st		(r7+1),r4
		st		(r7+2),r5
		st		(r7+3),r6
		; 	if ((cluster^FATvar.last_cluster)&0xffffff80) {
		ldi		r3,fatbas
		ld		r4,(r3+last_cluster-fatbas)
		ld		r5,(r3+last_cluster-fatbas+1)
		xor		r5,r5,r2
		xor		r4,r4,r1
		ldpc	r6
		word	0xff80
		and		r4,r4,r6
		or		r4,r4,r5
		jz		nxc1
; 		// read new FAT chunk
; 		FATvar.last_cluster=cluster;
		st		(r3+last_cluster-fatbas),r1
		st		(r3+last_cluster-fatbas+1),r2
; 		if (rdsector(FATvar.fat1_begin+(cluster>>7),FATbuf)) return 0;
		; (cluster>>7)
		ldi		r6,7
nxc2:	shr		r2,r2
		rorc	r1,r1
		subi	r6,1
		jnz		nxc2
		; +fat1_begin
		ld		r4,(r3+fat1_begin-fatbas)
		ld		r5,(r3+fat1_begin-fatbas+1)
		add		r1,r1,r4
		adc		r2,r2,r5
		ldpc	r3
		word	fatbuf
		jal		sd_rd_block
		or		r0,r0,r0
		jnz		nxc99
		ldi		r3,fatbas
		ld		r1,(r3+last_cluster-fatbas)
; 	pi=(u32 *)FATbuf; 	return pi[cluster&127];		
nxc1:	ldi		r5,127
		and		r5,r1,r5
		add		r5,r5,r5
		ldpc	r6
		word	fatbuf	; byte swapped area
		add		r5,r5,r6
		ld		r1,(r5)
		ld		r2,(r5+1)
		rori	r1,r1,8
		rori	r2,r2,8
		
nxc99:	ld		r3,(r7+0)
		ld		r4,(r7+1)
		ld		r5,(r7+2)
		ld		r6,(r7+3)
		addi	r7,4
		jind	r6

;--------------------------------------
;--------------------------------------
; rdfile
; Parameters:	R3 : pointer to buffer
; Returns:		R0 : bytes read (512 unless EOF)
;				R3+= 256
; Modiffies:	R1, R2

rdfile:	subi	r7,5
		st		(r7+0),r3
		st		(r7+1),r4
		st		(r7+2),r5
		st		(r7+3),r6

		ldi		r5,fatbas
;	// next sector
;	nc=FATvar.filecluster;
;	ns=FATvar.filesector+1;
;	i=1<<FATvar.l2spc;	// i = sectors per cluster
		ld		r1,(r5+filecluster-fatbas)
		ld		r2,(r5+filecluster-fatbas+1)
		ld		r4,(r5+filesector-fatbas)
		addi	r4,1
		ld		r6,(r5+spc-fatbas)
;if (ns>=spc) {ns=0; nc=next_cluster32(nc);}
		sub		r6,r4,r6
		jnc		rdf1
		ldi		r4,0
		jal		next_cluster
;	if(FATvar.filecluster>=0x0ffffff0) return 0; //EOF
rdf1:	st		(r7+4),r4	; save ns
		ldpc	r0
		word	0x0FFF
		ld		r6,(r5+filecluster-fatbas+1)
		sub		r0,r0,r6
		jnz		rdf2
		ldpc	r0
		word	0xFFF0
		ld		r6,(r5+filecluster-fatbas)
		and		r6,r6,r0
		sub		r0,r6,r0
		jnz		rdf2
		ldi		r0,0
		jr		rdf99
; if (buf==FATbuf) FATvar.last_cluster=-1; 		
rdf2:	ldpc	r0
		word	fatbuf
		sub		r0,r3,r0
		jnz		rdf3
		ldpc	r0
		word	0xFFFF
		st		(r5+last_cluster-fatbas),r0
		st		(r5+last_cluster-fatbas+1),r0
;	i=(	FATvar.storage_begin+((FATvar.filecluster-2)<<FATvar.l2spc) ) + FATvar.filesector;
rdf3:	ld		r4,(r5+filecluster-fatbas)
		ld		r6,(r5+filecluster-fatbas+1)
		subi	r4,2
		sbci	r6,0
		ld		r0,(r5+l2spc-fatbas)
rdf4:	add		r4,r4,r4
		adc		r6,r6,r6
		subi	r0,1
		jnz		rdf4
		ld		r0,(r5+storage_begin-fatbas)
		add		r4,r4,r0
		ld		r0,(r5+storage_begin-fatbas+1)
		adc		r6,r6,r0
		ld		r0,(r5+filesector-fatbas)
		add		r4,r4,r0
		adci	r6,0
;	FATvar.filesector=ns;
;	FATvar.filecluster=nc;
		ld		r0,(r7+4)
		st		(r5+filesector-fatbas),r0
		st		(r5+filecluster-fatbas),r1
		st		(r5+filecluster-fatbas+1),r2
;	if (rdsector(i,buf)) return 0;
		or		r1,r4,r4
		or		r2,r6,r6
		jal		sd_rd_block
		or		r0,r0,r0
		jnz		rdf99
;	if(FATvar.filelen>512) {FATvar.filelen-=512; return 512;}
;	else {i=FATvar.filelen; FATvar.filelen=0; return i; }
		ldpc	r0
		word	512
		ld		r4,(r5+filelen-fatbas)
		sub		r4,r4,r0
		ld		r6,(r5+filelen-fatbas+1)
		sbci	r6,0
		jpl		rdf5
		ld		r0,(r5+filelen-fatbas)
		ldi		r4,0
		ldi		r6,0
rdf5:	st		(r5+filelen-fatbas),r4
		st		(r5+filelen-fatbas+1),r6
		
rdf99:	ld		r3,(r7+0)
		ld		r4,(r7+1)
		ld		r5,(r7+2)
		ld		r6,(r7+3)
		addi	r7,5
		jind	r6	

;-----------------------------------------------
;       SPI
;-----------------------------------------------
;	SPI transfers (16 / 8 bits)
; Parameters:	R0: data to send
; Returns:		R0: data received
; Modiffies:	no other register

SPIxfer16:
		subi	r7,1
		st		(r7+0),r1
		ldi		r1,IOBASE
		st		(r1+OSPIDAT16),r0
		jr		bspx1
SPIxfer:subi	r7,1
		st		(r7+0),r1
		ldi		r1,IOBASE
		st		(r1+OSPIDAT),r0
bspx1:	ld		r0,(r1+OPFLAGS)
		andi	r0,SPIBUSY
		jnz		bspx1
		ld		r0,(r1+OSPIDAT16)
		ld		r1,(r7+0)
		addi	r7,1
		jind	r6			; retornamos	

;-----------------------------------------------
;       position to video address
;-----------------------------------------------
; Parameters:	pos: file selected
; Returns:		R2:	 byte pointer to video
; Modiffies:	none

pos2video:	; R2 = starting screen address for current position
		subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6

		ldi		r2,pos
		ld		r1,(r2)
		andi	r1,15
		rori	r1,r1,11
		ld		r2,(r2)
		andi	r2,16
		jz		pstv1
		addi	r1,15
pstv1:	ldpc	r2
		word	video*2+225
		add		r2,r2,r1

		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
;-----------------------------------------------
;       clear line at pos
;-----------------------------------------------
; Parameters:	pos: file selected
; Returns:		none
; Modiffies:	none

clrpos:	subi	r7,3
		st		(r7+0),r0
		st		(r7+1),r1
		st		(r7+2),r6
		jal		pos2video
		ldi		r0,' '
		ldi		r1,15
clps1:	jal		wrbyte
		subi	r1,1
		jnz		clps1
		ld		r0,(r7+0)
		ld		r1,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6
		
;-----------------------------------------------
;       write a Byte
;-----------------------------------------------
; Parameters:	R0: data
;               R2: byte address (=mem_address *2 +odd/even)
; Returns:		none
; Modiffies:	R2++
; Notes:		Big endian memory assumed
wrbyterem:
		subi	r7,4
		st		(r7+0),r0
		or		r0,r0,r3
		jr		wrb0
wrbyte:	subi	r7,4
		st		(r7+0),r0
wrb0:	st		(r7+1),r1
		st		(r7+2),r2
		st		(r7+3),r6
		shr		r2,r2
		ld		r6,(r2)		
		jc		vbl1
		andi	r6,0xff
		rori	r0,r0,8
		jr		vbl2
vbl1:	ldpc	r1
		word	0xFF00
		and		r6,r6,r1
vbl2:	or		r0,r0,r6
		st		(r2),r0

		ld		r0,(r7+0)
		ld		r1,(r7+1)
		ld		r2,(r7+2)
		addi	r2,1
		ld		r6,(r7+3)
		addi	r7,4
		jind	r6

;-----------------------------------------------
;       read a byte 
;-----------------------------------------------
; Parameters:	R1: byte address (=mem_address *2 +odd/even)
; Returns:		R0: byte read
; Modiffies:	R1++
; Notes:		Big endian memory assumed

rdbyte:	shr		r0,r1
		ld		r0,(r0)
		jc		rdb1
		rori	r0,r0,8
rdb1:	andi	r0,0xff
		addi	r1,1
		jind	r6

;-----------------------------------------------
;       strcpy
;-----------------------------------------------
; Parameters:	R0: max number of bytes to copy
;				R1: source byte pointer
;				R2: destination byte pointer
; Returns:		none
; Modiffies:	R1+=strlen  R2+=strlen
; Notes:        a zero terminator is added to destination so,
;				the total number of bytes written can be up to
;				R0+1
strncpy:
		subi	r7,3
		st		(r7+0),r0
		st		(r7+1),r5
		st		(r7+2),r6
		or		r5,r0,r0
stcp1:	jal		rdbyte	
		or		r0,r0,r0
		jz		stcp2
		jal		wrbyte
		subi	r5,1
		jnz		stcp1
		
stcp2:	;ldi		r0,0	; zero terminator
		;jal		wrbyte
		;subi		r2,1	; point to terminator
		ld		r0,(r7+0)
		ld		r5,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6

;-----------------------------------------------
;       bytecpy
;-----------------------------------------------
; Parameters:	R0: number of bytes to copy
;				R1: source byte pointer
;				R2: destination byte pointer
; Returns:		none
; Modiffies:	R1+=R0  R2+=R0
;bytecpy:
;;		subi	r7,3
;		st		(r7+0),r0
;		st		(r7+1),r5
;		st		(r7+2),r6
;		or		r5,r0,r0
;bycp1:	jal		rdbyte	
;		jal		wrbyte
;		subi	r5,1
;		jnz		bycp1
;	
;		ld		r0,(r7+0)
;		ld		r5,(r7+1)
;		ld		r6,(r7+2)
;		addi	r7,3
;		jind	r6

bytecpy:
		subi	r7,4
		st		(r7+0),r3
		ldi		r3,0
		jr		bycpr0
		
;-----------------------------------------------
;       bytecpyrem
;-----------------------------------------------
; Parameters:	R0: number of bytes to copy
;				R1: source byte pointer
;				R2: destination byte pointer
;				R3: remark value (0x80 or 0)
; Returns:		none
; Modiffies:	R1+=R0  R2+=R0
bytecpyrem:
		subi	r7,4
		st		(r7+0),r3
bycpr0:	st		(r7+1),r0
		st		(r7+2),r5
		st		(r7+3),r6
		or		r5,r0,r0
bycpr1:	jal		rdbyte	
		or		r0,r0,r3
		jal		wrbyte
		subi	r5,1
		jnz		bycpr1
	
		ld		r3,(r7+0)
		ld		r0,(r7+1)
		ld		r5,(r7+2)
		ld		r6,(r7+3)
		addi	r7,4
		jind	r6

		
;-----------------------------------------------
; getkey:		Wait for key pressed
;-----------------------------------------------
; Parameters:	none
; Returns:		R0: scancode
; Modiffies:	none
	
getkeysc:		; get scancode
		ldi		r1,IOBASE
gksc1: 	ld		r0,(r1+OPFLAGS)
		andi	r0,KVAL
		jz		gksc1
		ld		r0,(r1+OKEYB)
		jind	r6
 
getkey:	subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6
		jr		gk1
gk0:	jal		getkeysc
gk1:	jal		getkeysc
		cmpi	r0,0xF0 ; filter released keys out
		jz		gk0
		cmpi	r0,0xE0	; filter extended scancodes out
		jz		gk1
		
		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6

		
txterrsd:	asczbe "\xa0Error:  SDinit\xa0"
txterrfat:	asczbe "\xa0Error: FATinit\xa0"
txterrlin:	asczbe "\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0"
txto1:		asczbe "Floppyton-II J. Arias (2023)"
txto2:		asczbe "GUS16-V6 in the Apple"
txttrk:		asczbe "FloppyGUS-II wrote this track"
endtxt:

;----------------------------------------------
; Send data to UART
; Parameters:	R0: data to send
; Returns:		none
; Modiffies:	none
;
putch:	subi	r7,1
		st		(r7),r6
ptch1:	ldi		r6,IOBASE		; wait while TX busy
		ld		r6,(r6+OPFLAGS)		
		andi	r6,UARTBUSY
		jnz		ptch1
		ldi		r6,IOBASE		
		st		(r6+OUARTDAT),r0	; send data
		ld		r6,(r7)
		addi	r7,1
		jind	r6		


;----------------------------------------------
; Get data from UART (blocking)
; Parameters:	none
; Returns:		R0: data received
; Modiffies:	none
;
;getch:	ldi		r0,IOBASE	
;		ld		r0,(r0+OPFLAGS)	; wait until data valid
;		tsti	r0,UARTDV
;		jz		getch
;		ldi		r0,IOBASE
;		ld		r0,(r0+OUARTDAT)	; get the data
;		jind	r6

;-----------------------------------------------
;       puts 
;-----------------------------------------------
; Parameters:	R0: byte pointer
; Returns:		R0:	byte pointer updated R0+=strlen(string)
; Modiffies:	none
;
;puts:	subi	r7,r7,2
;		st		(r7+0),r1
;		st		(r7+1),r6
;		or		r1,r0,r0
;pts1:	adpc	r6,1
;		jr		rdbyte	
;		or		r0,r0,r0
;		jz		pts2	
;		adpc	r6,1
;		jr		putch
;		jr		pts1
;pts2:	or		r0,r1,r1
;		ld		r1,(r7+0)
;		ld		r6,(r7+1)
;		addi	r7,r7,2
;		jind	r6

;-----------------------------------------------
;		print hexadecimal number
;-----------------------------------------------
; Parameters:	R0:	data to print
; Returns:		none
; Modiffies:	no register
;
;-----------------------------------------------
;		print hexadecimal number
;-----------------------------------------------
; Parameters:	R0:	data to print
; Returns:		none
; Modiffies:	no register

prthex:	subi	r7,4
		st		(r7+0),r0
		st		(r7+1),r4
		st		(r7+2),r5
		st		(r7+3),r6
		or		r4,r0,r0
		ldi		r5,4
phl1:	rori	r4,r4,12
		or		r0,r4,r4
		andi	r0,15
		cmpi	r0,10
		jnc		phl2
		addi	r0,7
phl2:	addi	r0,'0'
		jal		putch
		subi	r5,1
		jnz		phl1

		ld		r0,(r7+2)	; R5
		jal		putch
		
		ld		r0,(r7+0)
		ld		r4,(r7+1)
		ld		r5,(r7+2)
		ld		r6,(r7+3)
		addi	r7,4
		jind	r6

;-----------------------------------------------
;		dump DMA buffer
;-----------------------------------------------
; Parameters:	none
; Returns:		none
; Modiffies:	no register
;
;dumpbuf:
;		subi	r7,r7,5
;		st		(r7+0),r0
;		st		(r7+1),r3
;		st		(r7+2),r4
;		st		(r7+3),r5
;		st		(r7+4),r6
;		
;		ldpc	r3
;		word	dmabuf
;		ldpc	r4
;		word	13*256
;dbu1:	ld		r0,(r3)
;		addi	r3,r3,1
;		ldi		r5,' '
;		tsti	r3,7
;		jnz		dbu2
;		ldi		r5,10
;dbu2:	adpc	r6,1
;		jr		prthex
;		subi	r4,r4,1
;		jnz		dbu1
;		
;		ld		r0,(r7+0)
;		ld		r3,(r7+1)
;		ld		r4,(r7+2)
;		ld		r5,(r7+3)
;		ld		r6,(r7+4)
;		addi	r7,r7,5
;		jind	r6


;-------- End of code
endcode:

;-------- special areas at memory top --------
		org	0x900
dmabuf:
fatbuf:		;block 512 bytes ; part of the File Allocation Table
filedata=	fatbuf+256	;block 32*32 bytes ; copy of "decent" files
		org 11*512
video:
;-------- end of memory -------
		org	12*512
stack:

