///////////////////////////////////////////////////////
// Apple II computer recreated
//  by Jesus Arias (2023)
//
//-------------------------------------------------------------------
//-- Test Bench
//-------------------------------------------------------------------

module tb();	// TOP module for simulations

//-- Input stimula
reg clk;
reg rxd;
reg reset;

//-- Internal nodes
wire [15:0]xa;
wire [15:0]xdo;
wire [15:0]xdi;
wire oe,we,ble,bhe;
wire sck, miso, mosi, fssb;
wire txd;

// Apple II instance
SYSTEM sys(.txd(txd), .clk4x(clk), .rxd(rxd), .reset(reset),
			.xwe(we), .xoe(oe),.xa(xa),.xdo(xdo),.xdi(xdi),
			.xble(ble), .xbhe(bhe),
			.sck(sck), .mosi(mosi), .miso(miso), .ss1(fssb), 
			.kclk(1'b1),.kdat(1'b1)
);

// External RAM (simulated here)
exRAM XRAM(.addr(xa), .data_in(xdo), .data_out(xdi), .oe(oe), .we(we),
			.bhe(bhe), .ble(ble) );
			
// External Flash SPI (simulated here)
FLASHSPI flashmem (
	.sck(sck), .mosi(mosi), .miso(miso), .ssb(fssb)
);

// Clock cycle of 10 units (simulation time/10 = cycles)
always #5 clk=~clk;

initial begin
	//-- output file with waveforms
	$dumpfile("tb.vcd");
	$dumpvars(0, tb);

	clk=0;
	reset=1;
	rxd=1;
	#50 reset=0;
	#80000 rxd=1;
	#1080 rxd=0;	// start
	#1080 rxd=1;	// d0
	#1080 rxd=0;	// d1
	#1080 rxd=1;	// d2
	#1080 rxd=0;	// d3
	#1080 rxd=0;	// d4
	#1080 rxd=1;	// d5
	#1080 rxd=0;	// d6
	#1080 rxd=1;	// d7
	#1080 rxd=1;	// stop
	
	#1080 rxd=0;	// start
	#1080 rxd=1;	// d0
	#1080 rxd=1;	// d1
	#1080 rxd=1;	// d2
	#1080 rxd=0;	// d3
	#1080 rxd=0;	// d4
	#1080 rxd=1;	// d5
	#1080 rxd=0;	// d6
	#1080 rxd=0;	// d7
	#1080 rxd=1;	// stop

	//# 319 $display("FIN de la simulacion");
	# 10000 $finish;
end

endmodule

//----------------------------------------------------------------------------
//-- External Asynchronous RAM
//----------------------------------------------------------------------------

module exRAM (        
    input wire [15:0] addr,      
    input wire [15:0] data_in,   
    output wire [15:0] data_out, 
	input wire	oe,
	input wire  we,
	input wire bhe,
    input wire ble
);
  //-- Memory (2 byte lanes)
  reg [7: 0] ramh [0: 16'hffff];
  reg [7: 0] raml [0: 16'hffff];

  //-- write
  always @(negedge we) begin
    if (~ble) raml[addr] <= data_in[7:0];
    if (~bhe) ramh[addr] <= data_in[15:8];
  end
  //-- asynchronous read
  assign data_out = {ramh[addr],raml[addr]};
endmodule

/////////////////////////////////////////////////////////////////////////////////////////////////////////
//----------------------------------------------------------------------------
//-- External SPI
//----------------------------------------------------------------------------

module FLASHSPI(
	input sck,
	input mosi,
	output miso,
	input ssb
);

reg [22:0]cnt=0;
always @(posedge sck or posedge ssb) if (ssb) cnt<=0; else cnt<=cnt+1;

reg [23:0]sh;
always @(posedge sck) sh<={sh[22:0],mosi};
reg [7:0]cmd;
reg[23:0]addr;
always @(posedge sck or posedge ssb) if (ssb) cmd<=0; else if (cnt==8) cmd<=sh[7:0];

always @(negedge sck or posedge ssb)
  if (ssb) addr<=0; else 
	if ((cmd==3)&(cnt==32)) addr<=sh; else
	if ((cnt>31) & (bit==0) & (cmd==3)) begin
		addr<=addr+1;
	end
reg [2:0]bit;
always @(negedge sck) bit<=~cnt[2:0];
wire [19:0]wrapaddr=addr[19:0];
wire [7:0]do=mem[wrapaddr]; 
assign miso=(cnt<32)? 1'b0 : do[bit];

reg [7:0]mem[0:(1<<20)-1];

initial begin
  $readmemh("FlashROM.hex", mem);
end
endmodule


